@@ -23,6 +23,7 @@ from ops.charm import (
     CharmBase,
     CharmMeta,
     CharmEvents,
+    ContainerMeta,
 )
 from ops.framework import Framework, EventSource, EventBase
 from ops.model import Model, _ModelBackend
@@ -235,6 +236,47 @@ storage:
             'StorageAttachedEvent',
         ])
 
+    def test_workload_events(self):
+
+        class MyCharm(CharmBase):
+            def __init__(self, *args):
+                super().__init__(*args)
+                self.seen = []
+                self.count = 0
+                for workload in ('container-a', 'containerb'):
+                    # Hook up relation events to generic handler.
+                    self.framework.observe(
+                        self.on[workload].pebble_ready,
+                        self.on_any_pebble_ready)
+
+            def on_any_pebble_ready(self, event):
+                self.seen.append(type(event).__name__)
+                self.count += 1
+
+        # language=YAML
+        self.meta = CharmMeta.from_yaml(metadata='''
+name: my-charm
+containers:
+  container-a:
+  containerb:
+''')
+
+        charm = MyCharm(self.create_framework())
+
+        self.assertIn('container_a_pebble_ready', repr(charm.on))
+        self.assertIn('containerb_pebble_ready', repr(charm.on))
+
+        charm.on['container-a'].pebble_ready.emit(
+            charm.framework.model.unit.get_container('container-a'))
+        charm.on['containerb'].pebble_ready.emit(
+            charm.framework.model.unit.get_container('containerb'))
+
+        self.assertEqual(charm.seen, [
+            'PebbleReadyEvent',
+            'PebbleReadyEvent'
+        ])
+        self.assertEqual(charm.count, 2)
+
     @classmethod
     def _get_action_test_meta(cls):
         # language=YAML
@@ -329,3 +371,17 @@ start:
 
     def test_action_event_defer_fails(self):
         self._test_action_event_defer_fails('action')
+
+    def test_containers(self):
+        meta = CharmMeta.from_yaml("""
+name: k8s-charm
+containers:
+  test1:
+    k: v
+  test2:
+    k: v
+""")
+        self.assertIsInstance(meta.containers['test1'], ContainerMeta)
+        self.assertIsInstance(meta.containers['test2'], ContainerMeta)
+        self.assertEqual(meta.containers['test1'].name, 'test1')
+        self.assertEqual(meta.containers['test2'].name, 'test2')
